/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#if !defined (BLOSUM55H)
#define BLOSUM55H

static float blosum55L[21][21] = {
/*    C           G           A           S           T       
      N           D           E           Q           K       
      R           H           W           Y           F       
      V           I           L           M           P       */
{1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.,1.},
{1.,   18.4929,    0.4820,    0.9358,    0.7982,    0.8542, 
   0.5091,    0.3796,    0.3684,    0.4257,    0.3945, 
   0.3741,    0.4424,    0.3980,    0.4738,    0.5288, 
   0.7932,    0.6206,    0.6297,    0.6250,    0.4566, },
{1.,   0.4820,    6.2571,    1.0626,    0.9552,    0.6242, 
   0.9399,    0.6889,    0.5181,    0.5922,    0.6288, 
   0.4883,    0.5669,    0.4613,    0.4252,    0.3801, 
   0.3807,    0.3423,    0.3394,    0.4625,    0.5445, },
{1.,   0.9358,    1.0626,    3.4803,    1.4164,    0.9949, 
   0.6758,    0.6092,    0.7633,    0.7849,    0.7556, 
   0.6499,    0.6116,    0.4321,    0.6193,    0.5368, 
   0.9641,    0.6856,    0.6245,    0.7915,    0.7217, },
{1.,   0.7982,    0.9552,    1.4164,    3.4438,    1.5338, 
   1.1974,    0.9210,    0.9033,    0.9952,    0.9288, 
   0.7932,    0.7708,    0.3715,    0.5999,    0.5037, 
   0.6227,    0.5001,    0.4707,    0.6335,    0.7557, },
{1.,   0.8542,    0.6242,    0.9949,    1.5338,    4.0295, 
   1.0600,    0.7658,    0.7785,    0.8287,    0.8147, 
   0.7096,    0.5956,    0.4515,    0.6345,    0.5267, 
   0.9912,    0.8294,    0.6928,    0.8415,    0.7161, },
{1.,   0.5091,    0.9399,    0.6758,    1.1974,    1.0600, 
   6.0898,    1.5462,    0.9337,    1.0075,    1.0029, 
   0.8531,    1.3292,    0.3245,    0.5784,    0.3936, 
   0.4192,    0.4145,    0.3735,    0.5374,    0.5740, },
{1.,   0.3796,    0.6889,    0.6092,    0.9210,    0.7658, 
   1.5462,    6.6405,    1.6828,    0.9207,    0.8675, 
   0.6781,    0.8102,    0.2876,    0.4482,    0.3004, 
   0.4099,    0.3735,    0.3460,    0.4089,    0.6582, },
{1.,   0.3684,    0.5181,    0.7633,    0.9033,    0.7785, 
   0.9337,    1.6828,    4.8828,    1.7742,    1.3545, 
   0.9686,    0.8836,    0.4625,    0.5939,    0.3897, 
   0.4757,    0.3902,    0.4328,    0.5222,    0.7573, },
{1.,   0.4257,    0.5922,    0.7849,    0.9952,    0.8287, 
   1.0075,    0.9207,    1.7742,    5.3755,    1.4895, 
   1.3742,    1.2329,    0.5986,    0.7149,    0.3782, 
   0.5162,    0.4399,    0.5397,    0.9218,    0.7168, },
{1.,   0.3945,    0.6288,    0.7556,    0.9288,    0.8147, 
   1.0029,    0.8675,    1.3545,    1.4895,    4.2210, 
   2.0750,    0.8986,    0.4228,    0.6010,    0.4005, 
   0.4950,    0.4263,    0.4529,    0.6345,    0.7631, },
{1.,   0.3741,    0.4883,    0.6499,    0.7932,    0.7096, 
   0.8531,    0.6781,    0.9686,    1.3742,    2.0750, 
   5.8763,    0.9862,    0.4701,    0.6613,    0.4499, 
   0.4828,    0.4147,    0.5319,    0.6453,    0.4905, },
{1.,   0.4424,    0.5669,    0.6116,    0.7708,    0.5956, 
   1.3292,    0.8102,    0.8836,    1.2329,    0.8986, 
   0.9862,    12.1409,    0.4762,    1.5618,    0.7230, 
   0.3761,    0.3898,    0.4740,    0.6520,    0.5508, },
{1.,   0.3980,    0.4613,    0.4321,    0.3715,    0.4515, 
   0.3245,    0.2876,    0.4625,    0.5986,    0.4228, 
   0.4701,    0.4762,    33.3867,    2.1342,    1.4160, 
   0.4226,    0.4707,    0.5568,    0.6553,    0.3346, },
{1.,   0.4738,    0.4252,    0.6193,    0.5999,    0.6345, 
   0.5784,    0.4482,    0.5939,    0.7149,    0.6010, 
   0.6613,    1.5618,    2.1342,    7.9348,    2.5840, 
   0.6929,    0.7280,    0.8146,    0.8354,    0.4244, },
{1.,   0.5288,    0.3801,    0.5368,    0.5037,    0.5267, 
   0.3936,    0.3004,    0.3897,    0.3782,    0.4005, 
   0.4499,    0.7230,    1.4160,    2.5840,    7.2030, 
   0.8065,    0.9727,    1.2440,    1.0662,    0.3350, },
{1.,   0.7932,    0.3807,    0.9641,    0.6227,    0.9912, 
   0.4192,    0.4099,    0.4757,    0.5162,    0.4950, 
   0.4828,    0.3761,    0.4226,    0.6929,    0.8065, 
   3.3424,    2.3725,    1.3066,    1.2165,    0.4838, },
{1.,   0.6206,    0.3423,    0.6856,    0.5001,    0.8294, 
   0.4145,    0.3735,    0.3902,    0.4399,    0.4263, 
   0.4147,    0.3898,    0.4707,    0.7280,    0.9727, 
   2.3725,    3.6371,    1.6668,    1.4182,    0.4620, },
{1.,   0.6297,    0.3394,    0.6245,    0.4707,    0.6928, 
   0.3735,    0.3460,    0.4328,    0.5397,    0.4529, 
   0.5319,    0.4740,    0.5568,    0.8146,    1.2440, 
   1.3066,    1.6668,    3.5764,    1.9038,    0.4127, },
{1.,   0.6250,    0.4625,    0.7915,    0.6335,    0.8415, 
   0.5374,    0.4089,    0.5222,    0.9218,    0.6345, 
   0.6453,    0.6520,    0.6553,    0.8354,    1.0662, 
   1.2165,    1.4182,    1.9038,    5.8803,    0.4742, },
{1.,   0.4566,    0.5445,    0.7217,    0.7557,    0.7161, 
   0.5740,    0.6582,    0.7573,    0.7168,    0.7631, 
   0.4905,    0.5508,    0.3346,    0.4244,    0.3350, 
   0.4838,    0.4620,    0.4127,    0.4742,    11.0494, }
};

#endif

