/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/**************************** binomial.h - ********************************
  Binomial distribution.
 *************************************************************************/
#if !defined(_WATCHGIBBS)
#define _WATCHGIBBS
#include "stdinc.h"
#include "afnio.h"
#include "random.h"
#include "histogram.h"

/************************ Binomial Data Structure  ************************/
typedef struct {
        float		**map;		/** random seed for binomial dist **/
        float		bestmap;	
	Int4		N;		/** number of reads **/
	Int4		*n;		/** number of specific read **/
	Int4		bestcol;
	Int4		lowcol,hicol;
	Int4		maxblk,maxcol;
} watchgibbs_type;

typedef watchgibbs_type *wg_type;

/******************************** private ********************************/

/********************************* PUBLIC ********************************/
wg_type MakeWatchGibbs(Int4 maxblk, Int4 maxcol);
void    PutWatchGibbs(FILE *fp, wg_type WG);
void    NilWatchGibbs(wg_type WG);
BooLean	AddWatchGibbs(float map, Int4 blk, Int4 col, wg_type WG);

/********************************* MACROS ********************************/

#endif

